/*
 * [Untouch.java]
 *
 * Summary: Untouch: Resets lastChangedDate of files if they have not really changed.
 *
 * Copyright: (c) 2003-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.7 2009-01-11 round all file timestamps to nearest milli so original/copied/untouched timestamps will match.
 *  1.8 2009-04-21 (-clear) option to clear history and set a current times as the revert-to times.
 *  1.9 2009-04-21 (-remove) option to remove untouch history i.e. untouch.digest and untouch.log files
 *  2.0 2009-05-04 now ignores .svn subversion directories.
 *  2.1 2009-06-05 add -help flag
 *  2.2 2011-01-16 fix bug stopping from Untouch rebuilding the digests after a remove.
 *  2.3 2011-11-07 add display of counts of various categories of files in the legend.
 *  2.4 2011-11-22 suppress log items for untouch.digest and untouch.log. It was just clutter.
 *  2.5 2013-04-17 allow slop in matching dates to account for Windows->Java loss of precision.
 *  2.6 2014-04-03 ensure all files we process have timestamps rounded to match exactly.
 */
package com.mindprod.untouch;

import com.mindprod.common18.EIO;
import com.mindprod.filter.AllButSVNDirectoriesFilter;
import com.mindprod.filter.OnlyFilesFilter;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.zip.Adler32;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import static com.mindprod.untouch.Changes.*;
import static java.lang.System.*;

/**
 * Untouch: Resets lastChangedDate of files if they have not really changed.
 * <p/>
 * times to avoid rounding.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 2.6 2014-04-03 ensure all files we process have timestamps rounded to match exactly.
 * @since 2003
 */
public final class Untouch
    {
    /**
     * Are file names case-sensitive?
     */
    private static final boolean CASE_SENSITIVE = false;

    /**
     * version of new untouch.digest file, current program file version
     */
    private static final int expectedFileVersion = 2;

    private static final int FIRST_COPYRIGHT_YEAR = 2003;

    /**
     * undisplayed copyright notice
     */
    @SuppressWarnings( { "UnusedDeclaration" } )
    private static final String EMBEDDED_COPYRIGHT =
            "Copyright: (c) 2003-2017 Roedy Green, Canadian Mind Products, http://mindprod.com";

    @SuppressWarnings( { "UnusedDeclaration" } )
    private static final String RELEASE_DATE = "2014-04-03";

    /**
     * name of application.
     */
    private static final String TITLE_STRING = "Untouch";

    /**
     * how to use Untouch
     */
    private static final String USAGE = "\nUntouch files, i.e. revert file dates to what they were previously if files " +
                                        "have not really changed.\n"
                                        + "Use: untouch.jar [ -flag1 -flag2 ... ] dir1 dir2 ...\n"
                                        + "possible flags:\n"
                                        + "-h = help\n"
                                        + "-c = clear history and start over with current file dates.\n"
                                        + "-f = force reversion to previous date, even if contents changed.\n"
                                        + "-l = log details of untouch activity.\n"
                                        + "-q = quiet. Suppress progress messages.\n"
                                        + "-r = remove untouch digests. Stop monitoring these dirs with untouch.\n"
                                        + "-s = process files in subdirectories too\n";

    /**
     * embedded version string.
     */
    private static final String VERSION_STRING = "2.6";

    /**
     * FilenameFilter that accepts only directories, except for .svn subversion directories. No files.
     */
    private static final AllButSVNDirectoriesFilter justDirs = new AllButSVNDirectoriesFilter();

    /**
     * FilenameFilter that accepts only files.
     */
    private static final OnlyFilesFilter justFiles = new OnlyFilesFilter();

    /**
     * version of the old untouch.digest file
     */
    private static int oldFileVersion = expectedFileVersion;

    /**
     * true if want history cleared first, as if there were no existing untouch.digest
     */
    private final boolean clear;

    /**
     * true if you want progress messages suppressed.
     */
    private final boolean quiet;

    /**
     * true if want human-readable untouch.log files generated.
     */
    private final boolean wantLogs;

    /**
     * use whacking big buffer for digester. This has to be fast. Instance rather than local to avoid reallocating it
     * over and over causing garbage collection.
     */
    private final byte[] ba = new byte[ 1024 * 128 ];

    /**
     * facts about files in directory now. Sorted in order by lower case file name.
     */
    private ArrayList<FileFacts> newFileFacts;

    /**
     * Previously recorded facts about files. Sorted in order by lower case file name. If empty will have [0], not
     * null.
     */
    private FileFacts[] oldFileFacts;

    /**
     * log file stored in the directory processed.
     */
    private PrintWriter log;

    /**
     * list of all filenames in directory, in lower case, alphabetical order.
     */
    private String[] filenames;

    /**
     * used by match to optimise searching. Must be initialised to 0 before matching on each directory.
     */
    private int matchPlace = 0;

    /**
     * constructor. Also used by ReplicatorSender and UntouchMindprod.
     *
     * @param quiet    true if you want progress messages suppressed.
     * @param wantLogs true if want human-readable untouch.log files generated.
     */
    public Untouch( boolean quiet, boolean wantLogs )
        {
        this( quiet, wantLogs, false /* no clear */  /* no remove */ );
        }

    /**
     * constructor. Also used by ReplicatorSender and UntouchMindprod.
     *
     * @param quiet    true if you want progress messages suppressed.
     * @param wantLogs true if want human-readable untouch.log files generated.
     * @param clear    true if want history cleared first.
     */
    public Untouch( boolean quiet, boolean wantLogs, boolean clear )
        {
        this.quiet = quiet;
        this.wantLogs = wantLogs;
        this.clear = clear;
        }

    /**
     * Display an explanation of the letter codes that explain what happened to the each file. Displayed on screen after
     * the dust settles, not in each untouch.log file. Also displays counts of files in each category.
     *
     * @param quiet    true if want progress messages suppressed.
     * @param wantLogs true if want untouch.log files generated.
     */
    private static void explainLegend( boolean quiet, boolean wantLogs )
        {
        if ( quiet )
            {
            return;
            }
        out.println();
        out.println( "         L E G E N D" );
        for ( Changes change : Changes.values() )
            {
            out.format( "%5d  %c %s", change.getCount(), change.getCode(), change.getDescription() );
            out.println();
            }
        out.println();
        if ( wantLogs )
            {
            out.println( "See untouch.log in each directory for details that scrolled off screen." );
            }
        out.println( TITLE_STRING + " " + VERSION_STRING + " finished" );
        }

    /**
     * Get canonical (fully qualified) directory name.
     *
     * @param dir Dir to get the name of
     *
     * @return String in exact case of actual dir name.
     */
    private static String getCanonicalDirectory( File dir )
        {
        return EIO.getCanOrAbsPath( dir );
        }

    /**
     * Process one directory, possibly the whole tree, possibly force. One directory name from the command line.
     *
     * @param directory name of directory.
     * @param force     true if want files restored to old dates whether changed or not.
     * @param tree      true if want the whole file tree processed.
     * @param quiet     true if want progress messages suppressed.
     * @param wantLogs  true if want untouch.log files generated.
     * @param clear     true if want history cleared first.
     * @param remove    true if want untouch history removed.
     */
    private static void processDirectory( String directory,
                                          boolean force,
                                          boolean tree,
                                          boolean quiet,
                                          boolean wantLogs,
                                          boolean clear,
                                          boolean remove )
        {
        final File dir = new File( directory );
        if ( !dir.isDirectory() )
            {
            return;// ignore unless it is a directory, would happen if *.* expanded by OS
            }
        final Untouch untouch = new Untouch( quiet, wantLogs, clear );
        if ( remove )
            {
            if ( tree )
                {
                untouch.removeAllUntouchHistoryInTree( dir );
                }
            else
                {
                untouch.removeAllUntouchHistoryInDir( dir );
                }
            }
        else if ( force )
            {
            if ( tree )
                {
                untouch.forceAllFilesInTree( dir );
                }
            else
                {
                untouch.forceAllFilesInDir( dir );
                }
            }
        else
            {
            if ( tree )
                {
                untouch.untouchAllFilesInTree( dir );
                }
            else
                {
                untouch.untouchAllFilesInDir( dir );
                }
            }
        }

    /**
     * Open the log file.
     *
     * @param dir Directory where to place the untouch.log file.
     */
    private void closeLog( File dir )
        {
        if ( !wantLogs )
            {
            final File stray = new File( dir, "untouch.log" );
            if ( stray.exists() )
                {
                //noinspection ResultOfMethodCallIgnored
                stray.delete();
                }
            }
        else if ( log != null )
            {
            log.close();
            }
        }

    /**
     * Compute a 32-bit digest of the file. If the file changes, the digest will almost surely change to. If the digest
     * is different, the file is surely different.
     *
     * @param file file to compute the digest of.
     *
     * @return Adlerian 32-bit digest.
     */
    private int computeDigest( File file )
        {
        final Adler32 digester = new Adler32();
        try
            {
            // O P E N
            FileInputStream fis = new FileInputStream( file );
            while ( true )
                {
                // R E A D
                int bytesRead = fis.read( ba, 0/* offset in ba */, ba.length );
                if ( bytesRead <= 0 )
                    {
                    /* -1=end of file, or 0=stall which should never happen */
                    break;
                    }
                else
                    {
                    digester.update( ba, 0/* offset */, bytesRead );
                    }
                } // end while
            // C L O S E
            fis.close();
            }
        catch ( IOException e )
            {
            // presumably we have no read access
            return 0;
            }
        return ( int ) digester.getValue();
        }

    /**
     * Get a list of filenames in the current directory, all lower case, in alphabetical order. leaves result in
     * filenames[].
     *
     * @param dir Directory where the files are
     */
    private void fetchFilenames( File dir )
        {
        filenames = dir.list( justFiles );
        if ( filenames == null )
            {
            // indicates no such directory
            filenames = new String[ 0 ];
            }
        if ( !CASE_SENSITIVE )
            {
            for ( int i = 0; i < filenames.length; i++ )
                {
                filenames[ i ] = filenames[ i ].toLowerCase();
                }
            }
        // sort alphabetically, all lower case, on mixed case, case-sensitive
        Arrays.sort( filenames );
        }

    /**
     * get previously recorded FileFacts about all files in this directory from the untouch.digest file. The List is in
     * alphabetical order. leaves result in oldFileFacts[].
     *
     * @param dir Directory where the filefacts untouch.digest is stored.
     */
    private void fetchOldFileFacts( File dir )
        {
        if ( clear )
            {
            // act is af existing untouch.digest were not there.
            oldFileFacts = new FileFacts[ 0 ];
            return;
            }
        try
            {
            // O P E N
            final FileInputStream fis = new FileInputStream( new File( dir, "untouch.digest" ) );
            final GZIPInputStream gzis = new GZIPInputStream( fis, 16 * 1024/* buffsize */ );
            final ObjectInputStream ois = new ObjectInputStream( gzis );
            // deal with complexity version 1 files don't have an explicit version number
            Object probe = ois.readObject();
            if ( probe instanceof Integer )
                {
                oldFileVersion = ( Integer ) probe;
                oldFileFacts = ( FileFacts[] ) ois.readObject();
                }
            else
                {
                oldFileVersion = 1;
                oldFileFacts = ( FileFacts[] ) probe;
                }
            // no big deal if old version is not same as expected.
            // C L O S E
            ois.close();
            }
        catch ( Exception e )
            {
            // expected on first use, untouch.digest will not exist
            if ( !quiet )
                {
                out.println( "No previous untouch.digest" );
                }
            oldFileFacts = new FileFacts[ 0 ];
            }
        }

    /**
     * Process one directory, forcing all files in it back to their old timestamps.
     *
     * @param dir Directory where the files are.
     */
    private void forceAllFilesInDir( File dir )
        {
        try
            {
            if ( !quiet )
                {
                out.println( "Untouch forcing " + getCanonicalDirectory( dir ) );
                }
            openLog( dir );
            fetchFilenames( dir );
            fetchOldFileFacts( dir );
            // We may have file facts for files that have been deleted, or there may be new files for which we do
            // not have file facts.
            // We are creating a new set of file facts for whatever files do exist now.
            // filenames.length is an upper bound.
            // We don't generate file facts for every file type.
            newFileFacts = new ArrayList<>( filenames.length );
            // process each file currently in the directory
            matchPlace = 0;
            for ( String filename : filenames )
                {
                // File must pass a gauntlet of conditions before
                // we consider setting its date back to what it was
                // previously.
                final File file = new File( dir, filename );
                if ( filename.equals( "untouch.digest" )
                     || filename.equals( "untouch.log" ) )
                    {
                    // quietly ignore them
                    continue;
                    }
                else if ( !file.canRead() )
                    {
                    // if we have not permission to compute digests, we can't
                    // touch this
                    // file.
                    tellStatus( IGNORING, filename );
                    continue;
                    }
                final long fileLength = file.length();
                // this is the MS filedate rounded.
                final long lastChangeDate = file.lastModified();
                // round so other files with same java/untouch time will match
                file.setLastModified( lastChangeDate );
                // procrastinate calculating digest since it is so time
                // consuming.
                // We don't want to do it unless we absolutely have to.
                FileFacts old = match( filename );
                if ( old == null )
                    {
                    // we have a brand new file.
                    tellStatus( NEW, filename );
                    newFileFacts.add( new FileFacts( filename,
                            fileLength,
                            lastChangeDate,
                            computeDigest( file ) ) );
                    // round off timestamp to the millisecond
                    file.setLastModified( lastChangeDate );
                    continue;
                    }
                // we have previous info on the file.
                if ( fileLength != old.fileLength )
                    {
                    /* the file has changed. Treat as a new file. */
                    tellStatus( LENGTH_CHANGED, filename );
                    newFileFacts.add( new FileFacts( filename,
                            fileLength,
                            lastChangeDate,
                            computeDigest( file ) ) );
                    file.setLastModified( old.lastChangeDate );
                    continue;
                    }
                if ( Math.abs( lastChangeDate - old.lastChangeDate ) <= FileFacts.SLOP )
                    {
                    tellStatus( UNCHANGED, filename );
                    // trust file is unchanged, even without computing digest
                    newFileFacts.add( old );
                    newFileFacts.add( old );
                    if ( oldFileVersion <= 1 )
                        {
                        // round off timestamp to the millisecond
                        file.setLastModified( lastChangeDate );
                        }
                    else
                        {
                        // no need to round off timestamp to the millisecond if already  done earlier by version 2+
                        }
                    continue;
                    }
                final int digest = computeDigest( file );
                if ( digest != old.digest )
                    {
                    // file changed. Treat as anew file.
                    tellStatus( CONTENTS_CHANGED, filename );
                    newFileFacts.add( new FileFacts( filename,
                            fileLength,
                            lastChangeDate,
                            digest ) );
                    file.setLastModified( old.lastChangeDate );
                    continue;
                    }
                // file has not changed, but timestamp did. Put timestamp back.
                // This case is the raison d'etre for untouch.
                newFileFacts.add( old );
                tellStatus( CONTENTS_UNCHANGED, filename );
                file.setLastModified( old.lastChangeDate );
                }
            saveNewFileFacts( dir );
            filenames = null;
            oldFileFacts = null;
            newFileFacts = null;
            closeLog( dir );
            }
        catch ( Exception e )
            {
            err.println();
            e.printStackTrace( err );
            err.println( "Untouch force failed: " + e.getMessage() );
            err.println();
            System.exit( 1 );
            }
        }

    /**
     * Force all files in the directory tree to revert to their old dates.
     *
     * @param dir Directory tree where the files are.
     */
    private void forceAllFilesInTree( File dir )
        {
        // process all files in this dir first
        forceAllFilesInDir( dir );
        // recursively process all subdirs in this directory
        final String[] allDirs = dir.list( justDirs );
        for ( String allDir : allDirs )
            {
            forceAllFilesInTree( new File( dir, allDir ) );
            }
        } // end forceAllFilesInTree

    /**
     * Find a matching FileFact for the given filename. Works with a linear merge-like search faster than binary
     * search. Lookups must be done in ascending
     * order. MatchPlace must be initialised to 0 before use on each new directory
     *
     * @param filename Filename to look up
     *
     * @return matching FileFact from previously, or null if one cannot be found.
     */
    private FileFacts match( String filename )
        {
        // needed outside the loop too
        final int n = oldFileFacts.length;
        for ( int i = matchPlace; i < n; i++ )
            {
            int comp = oldFileFacts[ i ].filename.compareTo( filename );
            if ( comp == 0 )
                {
                // we found it
                // next time start one past this one.
                matchPlace = i + 1;
                return oldFileFacts[ i ];
                }
            else if ( comp < 0 )
                {
                // we have not gone far enough.
                // keep looking
                }
            else
                {
                // we have gone too far.
                // next time start looking with this one.
                matchPlace = i;
                // we did not find it
                return null;
                }
            } // end for
        // keep returning null from now on for even higher names.
        matchPlace = n;
        return null;
        } // end match

    /**
     * Open the log file.
     *
     * @param dir Directory where to place the untouch.log file.
     */
    private void openLog( File dir )
        {
        if ( !wantLogs )
            {
            return;
            }
        try
            {
            final File logFile = new File( dir, "untouch.log" );
            log = EIO.getPrintWriter( logFile, 1024, EIO.UTF8 );
            }
        catch ( IOException e )
            {
            err.println( "Untouch: Trouble creating untouch.log file" );
            System.exit( 1 );
            }
        }

    /**
     * remove untouch.digest and untouch.log history from the current dir.
     *
     * @param dir the directory where the files are.
     */
    private void removeAllUntouchHistoryInDir( File dir )
        {
        if ( !quiet )
            {
            out.println( "Removing history " + getCanonicalDirectory( dir ) );
            }
        new File( dir, "untouch.digest" ).delete();
        new File( dir, "untouch.log" ).delete();
        }

    /**
     * remove all untouch history from this directory tree.
     *
     * @param dir Directory tree where the untouch history files are.
     */
    final void removeAllUntouchHistoryInTree( File dir )
        {
        // remove untouch.digest and untouch.log in this dir first
        removeAllUntouchHistoryInDir( dir );
        // recursively process all subdirs in this directory
        final String[] allDirs = dir.list( justDirs );
        for ( String allDir : allDirs )
            {
            removeAllUntouchHistoryInTree( new File( dir, allDir ) );
            }
        }

    /**
     * Save newFileFacts about the current directory back into an untouch.digest file.
     *
     * @param dir Directory where to story file facts, same as dir whose files it records.
     */
    private void saveNewFileFacts( File dir )
        {
        boolean mustWrite;
        // we can avoid rewriting untouch.digest if nothing has changed.
        if ( oldFileFacts.length == newFileFacts.size() )
            {
            // might be able to bypass work of saving.
            mustWrite = false;
            for ( int i = 0; i < oldFileFacts.length; i++ )
                {
                if ( !newFileFacts.get( i ).equals( oldFileFacts[ i ] ) )
                    {
                    mustWrite = true;
                    break;
                    }
                }
            }
        else
            {
            mustWrite = true;
            }
        if ( mustWrite )
            {
            try
                {
                // O P E N
                FileOutputStream fos = new FileOutputStream( new File( dir, "untouch.digest" ), false  /* append */ );
                GZIPOutputStream gzos = new GZIPOutputStream( fos, 16 * 1024   /* buffsize */ );
                ObjectOutputStream oos = new ObjectOutputStream( gzos );
                // W R I T E
                oos.writeObject( expectedFileVersion );
                oos.writeObject( newFileFacts.toArray( new FileFacts[ newFileFacts.size() ] ) );
                // C L O S E
                oos.close();
                }
            catch ( IOException e )
                {
                err.println( "Untouch: Trouble saving untouch.digest" );
                System.exit( 1 );
                }
            }
        }

    /**
     * Display and log status of what we did with a given file.
     *
     * @param status   code such as CONTENTS_CHANGED on what we did with that file.
     * @param filename name of file in question.
     */
    private void tellStatus( Changes status, String filename )
        {
        status.incrementCount();
        final StringBuilder sb = new StringBuilder( filename.length() + 2 );
        sb.append( status.getCode() );
        sb.append( ' ' );
        sb.append( filename );
        String message = sb.toString();
        if ( !quiet && status != UNCHANGED )
            {
            out.println( message );
            }
        if ( wantLogs && log != null )
            {
            log.println( message );
            }
        }

    /**
     * untouching all files in one directory that not really changed.
     *
     * @param dir the directory where the files are.
     */
    private void untouchAllFilesInDir( File dir )
        {
        try
            {
            if ( !quiet )
                {
                out.println( "Untouching " + getCanonicalDirectory( dir ) );
                }
            openLog( dir );
            fetchFilenames( dir );
            fetchOldFileFacts( dir );
            // We may have file facts for files that have been deleted, or there may be new files for which we do
            // not have file facts.
            // We are creating a new set of file facts for whatever files do exist now.
            // filenames.length is an upper bound.
            // We don't generate file facts for every file type.
            newFileFacts = new ArrayList<>( filenames.length );
            // process each file currently in the directory
            matchPlace = 0;
            for ( String filename : filenames )
                {
                // File must pass a gauntlet of conditions before
                // we consider setting its date back to what it was
                // previously.
                final File file = new File( dir, filename );
                if ( filename.equals( "untouch.digest" )
                     || filename.equals( "untouch.log" ) )
                    {
                    // quietly ignore them
                    continue;
                    }
                if ( !file.canRead() )
                    {
                    // if we have no permission to compute digests, we can't
                    // touch this
                    // file.
                    tellStatus( IGNORING, filename );
                    continue;
                    }
                final long fileLength = file.length();
                // this is the ms timestamp rounded
                final long lastChangeDate = file.lastModified();
                // round so other files with same java/untouch time will match
                file.setLastModified( lastChangeDate );
                // procrastinate calculating digest since it is so time
                // consuming.
                // We don't want to do it unless we absolutely have to.
                final FileFacts old = match( filename );
                if ( old == null )
                    {
                    // we have a brand new file.
                    tellStatus( NEW, filename );
                    newFileFacts.add( new FileFacts( filename,
                            fileLength,
                            lastChangeDate,
                            computeDigest( file ) ) );
                    // round off timestamp to the millisecond
                    file.setLastModified( lastChangeDate );
                    continue;
                    }
                // we have previous info on the file.
                if ( fileLength != old.fileLength )
                    {
                    /* the file has changed. Treat as a new file. */
                    tellStatus( LENGTH_CHANGED, filename );
                    newFileFacts.add( new FileFacts( filename,
                            fileLength,
                            lastChangeDate,
                            computeDigest( file ) ) );
                    // round off timestamp to the millisecond
                    //noinspection ResultOfMethodCallIgnored
                    file.setLastModified( lastChangeDate );
                    continue;
                    }
                if ( Math.abs( lastChangeDate - old.lastChangeDate ) <= FileFacts.SLOP )
                    {
                    tellStatus( UNCHANGED, filename );
                    // trust file is unchanged, even without computing digest
                    newFileFacts.add( old );
                    if ( oldFileVersion <= 1 )
                        {
                        // round off timestamp to the millisecond
                        file.setLastModified( lastChangeDate );
                        }
                    else
                        {
                        // no need to round off timestamp to the millisecond if already  done earlier by version 2+
                        }
                    continue;
                    }
                final int digest = computeDigest( file );
                if ( digest != old.digest )
                    {
                    // file changed. Treat as anew file.
                    tellStatus( CONTENTS_CHANGED, filename );
                    newFileFacts.add( new FileFacts( filename,
                            fileLength,
                            lastChangeDate,
                            digest ) );
                    // round off timestamp to the millisecond
                    file.setLastModified( lastChangeDate );
                    continue;
                    }
                // file has not changed, but timestamp did. Put timestamp back.
                // This case is the raison d'etre for untouch.
                newFileFacts.add( old );
                tellStatus( CONTENTS_UNCHANGED, filename );
                // will be rounded to millisecond
                file.setLastModified( old.lastChangeDate );
                } // end for
            // deleted files will not be copied over to newFileFacts
            // or have new entries generated.
            saveNewFileFacts( dir );
            filenames = null;
            oldFileFacts = null;
            newFileFacts = null;
            closeLog( dir );
            }
        catch ( Exception e )
            {
            err.println();
            e.printStackTrace( err );
            err.println( "Untouch failed: " + e.getMessage() );
            err.println();
            System.exit( 1 );
            }
        }

    /**
     * Marks files back to their previous dates if they have not changed content since the last time untouch was run.
     *
     * @param args one directory or directory tree name, with options:
     *             <ul>
     *             <li>-c or -clear to clear all history, and take the current file times as the new revert-to point
     *             <li>-f or -force to force all files back to their original timestamps, changed or not.</li>
     *             <li>-l or -log to generate untouch.log files of what was done.</li>
     *             <li>-q or -quiet to suppress progress messages</li>
     *             <li>-r or -remove to remove untouch history for specified directories</li>
     *             <li>-s or -subdir or -subdir or -t or -tree = processsubdirs too</li>
     *             </ul> switches must precede the dir name. You can put
     *             multiple dir names on the command line, but no individual file names. Without parameters, Untouch
     *             processes the current directory without subdirs.
     */
    public static void main( String[] args )
        {
        boolean force = false;
        boolean tree = false;
        boolean quiet = false;
        boolean wantLogs = false;
        boolean clear = false;
        boolean remove = false;
        int dirsProcessed = 0;
        for ( String arg : args )
            {
            if ( arg.equalsIgnoreCase( "-h" )
                 || arg.equalsIgnoreCase( "-help" ) )
                {
                out.println( USAGE );
                System.exit( 0 );
                }
            if ( arg.equalsIgnoreCase( "-c" )
                 || arg.equalsIgnoreCase( "-clear" ) )
                {
                clear = true;
                }
            else if ( arg.equalsIgnoreCase( "-f" )
                      || arg.equalsIgnoreCase( "-force" ) )
                {
                force = true;
                }
            else if ( arg.equalsIgnoreCase( "-l" ) || arg.equalsIgnoreCase(
                    "-log" ) )
                {
                wantLogs = true;
                }
            else if ( arg.equalsIgnoreCase( "-q" ) || arg.equalsIgnoreCase(
                    "-quiet" ) )
                {
                quiet = true;
                }
            else if ( arg.equalsIgnoreCase( "-r" ) || arg.equalsIgnoreCase(
                    "-remove" ) )
                {
                remove = true;
                }
            else if ( arg.equalsIgnoreCase( "-s" )
                      || arg.equalsIgnoreCase( "-sub" )
                      || arg.equalsIgnoreCase( "-subdir" )
                      || arg.equalsIgnoreCase( "-subdirs" )
                      || arg.equalsIgnoreCase( "-t" )
                      || arg.equalsIgnoreCase( "-tree" ) )
                {
                tree = true;
                }
            else if ( arg.startsWith( "-" ) )
                {
                err.println( "Unrecognised flag " + arg );
                err.println( USAGE );
                System.exit( 2 );
                }
            else
                {
                // was  a directory.
                processDirectory( arg, force, tree, quiet, wantLogs, clear, remove );
                dirsProcessed++;
                }
            } // end for
        // all done processing the command line.
        if ( dirsProcessed == 0 )
            {
            // default, process current dir.
            processDirectory( ".", force, tree, quiet, wantLogs, clear, remove );
            }
        // explain legend once at the very end
        explainLegend( quiet, wantLogs );
        } // end main

    /**
     * revert all files in the directory tree to their old dates.
     *
     * @param dir Directory tree where the files are.
     */
    public final void untouchAllFilesInTree( File dir )
        {
        // process all files in this dir first
        untouchAllFilesInDir( dir );
        // recursively process all subdirs in this directory
        final String[] allDirs = dir.list( justDirs );
        for ( String allDir : allDirs )
            {
            untouchAllFilesInTree( new File( dir, allDir ) );
            }
        } // end unTouchAllFilesInTree
    }
// end untouch
